import os
import requests

def download_example(image: str = "getting_started"):
    assert isinstance(image, str), "image must be a string"
    assert image in ["getting_started"], "image must be one of ['getting_started']"
    if image == "getting_started":
        url = "https://openslide.cs.cmu.edu/download/openslide-testdata/Aperio/CMU-1.svs"
        file_name = "getting_started.svs"
    else:
        raise ValueError("image must be one of ['getting_started']")
    assert not os.path.exists(file_name), f"{file_name} already exists"
    print(f"Downloading {image} from {os.path.dirname(url)}\nThis may take a few minutes...")
    response = requests.get(url)
    if response.status_code == 200:
        with open(f"{file_name}", "wb") as f:
            f.write(response.content)
    else:
        raise ConnectionError(f"Failed to download {image}")


from .._abc._model import Model
from ..process._process import Processor
from typing import List

class StainDeconvolver(Model):
    def __init__(self, name: str, preprocessor: Processor = None, postprocessor: Processor = None, out_stains = List[str]) -> None:
        super().__init__(name, preprocessor, postprocessor)
        if isinstance(out_stains, str):
            out_stains = [out_stains]
        assert isinstance(out_stains, list)
        self.out_stains = out_stains


from .._process import Process
from ...image._image import Image
from ...image._ubyte_image import UByteImage
import dask.array as da
from skimage import img_as_ubyte
from skimage.color import rgb2gray

class RGBToGrey(Process):
    def __init__(self) -> None:
        super().__init__('Convert an RGB UByte Image to a Greyscale UByte Image')

    def run(self, image: Image) -> Image:
        assert isinstance(image, UByteImage)
        assert image.ndim == 3
        assert image.shape[2] == 3
        assert image.channel_names == ["Red", "Green", "Blue"]
        img = image.image
        grey_img = da.map_blocks(lambda x: img_as_ubyte(rgb2gray(x)), img, drop_axis=2, dtype='uint8')
        grey_img = da.expand_dims(grey_img, axis=2)
        return UByteImage(grey_img, image.info, ["Grey"])


#   ---------------------------------------------------------------------------------
#   Copyright (c) Microsoft Corporation. All rights reserved.
#   Licensed under the MIT License. See LICENSE in project root for information.
#   ---------------------------------------------------------------------------------
"""
This is a configuration file for pytest containing customizations and fixtures.

In VSCode, Code Coverage is recorded in config.xml. Delete this file to reset reporting.
"""

from __future__ import annotations

from typing import List

import pytest
from _pytest.nodes import Item


def pytest_collection_modifyitems(items: list[Item]):
    for item in items:
        if "spark" in item.nodeid:
            item.add_marker(pytest.mark.spark)
        elif "_int_" in item.nodeid:
            item.add_marker(pytest.mark.integration)


@pytest.fixture
def unit_test_mocks(monkeypatch: None):
    """Include Mocks here to execute all commands offline and fast."""
    pass


from .._process import Process
from ...image._image import Image

class Rechunk(Process):
    def __init__(self, chunks: tuple) -> None:
        super().__init__(f'Rechunk image to {str(chunks)}')
        self.chunks = chunks

    def run(self, image: Image) -> Image:
        image.rechunk(self.chunks)
        return image


from ._gamma_contrast import GammaContrast
from ._linear_contrast import LinearContrast
from ._min_max_contrast import MinMaxContrast

__all__ = [
    'GammaContrast',
    'LinearContrast',
    'MinMaxContrast'
]


from ._mask import Mask
import dask.array as da
from dask.array import Array
import matplotlib.pyplot as plt
from numpy import ndarray
from typing import List

class BinaryMask(Mask):
    def __init__(self, image: Array, info: dict, channel_names: List[str]) -> None:
        super().__init__(image, info, channel_names)
        assert image.dtype == bool, "image must be a boolean array"

    def get_thumb(self, target_size: int = 512) -> ndarray:
        assert isinstance(target_size, int), "target_size must be an integer"
        assert target_size > 0, "target_size must be greater than 0"
        coarsen_factor = max([s // target_size for s in self.shape])
        if coarsen_factor == 0:
            coarsen_factor = 1
        image = self.image
        image = image.astype('float32')
        thumb = da.coarsen(da.mean, image, {0: coarsen_factor, 1: coarsen_factor}, trim_excess=True)
        return thumb.compute()

    def show_thumb(self, target_size: int = 512) -> None:
        thumb = self.get_thumb(target_size)
        plt.imshow(thumb, cmap='gray')
        plt.axis('off')
        plt.show()


from ..image._image import Image
from abc import ABC, abstractmethod

class Process(ABC):
    def __init__(self, name: str) -> None:
        assert isinstance(name, str), 'name must be a string'
        self.name = name

    @abstractmethod
    def run(self, image: Image) -> Image:
        pass

class Processor():
    def __init__(self) -> None:
        self.processes = []

    def add_process(self, process: Process) -> None:
        assert isinstance(process, Process), 'process must be a Process'
        self.processes.append(process)

    def run(self, image: Image) -> Image:
        assert isinstance(image, Image), 'image must be an Image'
        image = image
        for process in self.processes:
            image = process.run(image)
        return image

    def __repr__(self) -> str:
        if len(self.processes) == 0:
            ret_str = "Empty processor object"
        else:
            ret_str = "Processor object with processes:"
            for i, proc in enumerate(self.processes):
                ret_str += f"\n\t({i+1}) {proc.name}"
        return ret_str

    def _repr_html_(self) -> str:
        total_width = 400
        html = f' <div style="width: {total_width}px; background-color: #202020; padding: 20px; border-radius: 5px; box-shadow: 0 2px 4px rgba(0, 0, 0, 0.1);">'
        html += f'<h1>SCEMATK Processor Object</h1>'
        if len(self.processes) == 0:
            html += f'<p>Empty processor object</p>'
        else:
            html += f'<p>Processor object with processes:</p>'
            for i, proc in enumerate(self.processes):
                html += f'<p>({i+1}) {proc.name}</p>'
        html += '</div>'
        return html



from ._stain_deconvolver import StainDeconvolver
from ..process._process import Processor
from ..image._ubyte_image import UByteImage
import dask.array as da
from skimage.color import rgb2hed

class NativeSKImageStainDeconvolver(StainDeconvolver):
    def __init__(self, preprocessor: Processor = None, postprocessor: Processor = None, stain_type: str = "H&E") -> None:
        assert isinstance(stain_type, str), "stain_type must be a string"
        assert stain_type in ['H&E', 'H-DAB'], "stain_type must be either 'H&E' or 'H-DAB'"
        if stain_type == 'H&E':
            out_stains = ['Hematoxylin', 'Eosin']
        elif stain_type == 'H-DAB':
            out_stains = ['Hematoxylin', 'DAB']
        else:
            raise ValueError("stain_type must be either 'H&E' or 'H-DAB'")
        super().__init__("Stain deconvolver using scikit-image defaults", preprocessor, postprocessor, out_stains)
        self.fitted = True

    def fit(self, image: UByteImage) -> None:
        pass

    def run(self, image: UByteImage) -> UByteImage:
        image = self.preprocessor.run(image)
        assert isinstance(image, UByteImage), "image must be a UByteImage"
        assert image.channel_names == ['Red', 'Green', 'Blue'], "image must have channels ['Red', 'Green', 'Blue']"
        if self.out_stains == ['Hematoxylin', 'Eosin']:
            deconv_func = lambda x: rgb2hed(x)[:, :, :2]
        elif self.out_stains == ['Hematoxylin', 'DAB']:
            deconv_func = lambda x: rgb2hed(x)[:, :, [0, 2]]
        else:
            raise ValueError("out_stains must be either ['Hematoxylin', 'Eosin'] or ['Hematoxylin', 'DAB']")
        img = image.image
        chunks = image.image.chunks
        deconv_img = da.map_blocks(deconv_func, img, dtype='float64', chunks=(chunks[0], chunks[1], 2))
        deconv_img = deconv_img * 255
        deconv_img = deconv_img.astype('uint8')
        deconv_img = UByteImage(deconv_img, image.info, self.out_stains)
        deconv_image = self.postprocessor.run(deconv_img)
        return deconv_image


    def fit_and_run(self, image: UByteImage) -> UByteImage:
        return self.run(image)

    def _default_preprocessor(self) -> Processor:
        return Processor()

    def _default_postprocessor(self) -> Processor:
        return Processor()


from ._image import Image
from dask.array import Array
import dask.array as da
import matplotlib.pyplot as plt
import numpy as np
from numpy import ndarray
from typing import List

class UByteImage(Image):
    def __init__(self, image: Array, info: dict, channel_names: List[str]) -> None:
        super().__init__(image, info, channel_names)
        assert self.dtype == 'uint8', "image must be of type uint8"
        if self.ndim == 2:
            self.image = da.expand_dims(self.image, axis=-1)
        self.ndim = self.image.ndim
        self.shape = self.image.shape

    def get_thumb(self, target_size: int = 512) -> ndarray:
        assert isinstance(target_size, int), "target_size must be an integer"
        assert target_size > 0, "target_size must be greater than 0"
        coarsen_factor = max([s // target_size for s in self.shape])
        if coarsen_factor == 0:
            coarsen_factor = 1
        image = self.image
        thumb = da.coarsen(da.mean, image, {0: coarsen_factor, 1: coarsen_factor, 2:1}, trim_excess=True)
        thumb = thumb.astype('uint8')
        return thumb.compute()

    def show_thumb(self, target_size: int = 512) -> None:
        thumb = self.get_thumb(target_size)
        if self.shape[2] == 1:
            thumb = thumb.squeeze()
            cmap = 'gray'
        elif self.shape[2] == 2:
            thumb = np.pad(thumb, ((0, 0), (0, 0), (1, 0)), mode='constant', constant_values=0)
            cmap = None
        elif self.shape[2] == 3:
            cmap = None
        else:
            raise NotImplementedError("Only 1, 2, or 3 channels supported")
        plt.imshow(thumb, cmap=cmap)
        plt.axis('off')
        plt.show()


from ._image import Image
from dask.array import Array
from typing import List

class Mask(Image):
    def __init__(self, image: Array, info: dict, channel_names: List[str]) -> None:
        super().__init__(image, info, channel_names)
        assert self.ndim == 2, "mask must be 2D"